@@ -17,10 +17,11 @@
 #include <unistd.h>
 #include <sys/select.h>
 #include <sys/types.h>
+#include <sys/uio.h>
 #include <sys/wait.h>
 
-#include <arpa/inet.h>
 #include <netinet/in.h>
+#include <arpa/inet.h>
 
 #include "soup-dns.h"
 #include "soup-misc.h"
@@ -344,7 +345,9 @@ soup_dns_entry_unref (SoupDNSEntry *entry)
 {
 	if (!--entry->ref_count) {
 		g_free (entry->name);
-		soup_dns_free_hostent (entry->h);
+
+		if (entry->h)
+			soup_dns_free_hostent (entry->h);
 
 		if (entry->fd)
 			close (entry->fd);
@@ -656,6 +659,10 @@ gboolean
 soup_dns_entry_check_lookup (SoupDNSEntry *entry)
 {
 	check_hostent (entry, FALSE);
+
+	if (entry->resolved && entry->h == NULL)
+		uncache_entry (entry);
+
 	return entry->resolved;
 }
 
